from __future__ import annotations

from functools import lru_cache
from typing import Dict, Optional

from nebulo.sql.inspect import get_table_name
from sqlalchemy import text as sql_text
from sqlalchemy.engine import Engine
from sqlalchemy.sql.schema import Constraint

__all__ = ["populate_constraint_comment"]


SchemaName = str
TableName = str
ConstraintName = str
Comment = str

CommentMap = Dict[SchemaName, Dict[TableName, Dict[ConstraintName, Comment]]]


def populate_constraint_comment(engine: Engine, constraint: Constraint) -> None:
    """Adds SQL comments on a constraint to the SQLAlchemy constraint's
    Constraint.info['comment'] dictionary
    """

    schema: Optional[str] = constraint.table.schema
    table_name: str = get_table_name(constraint.table)
    constraint_name: Optional[str] = constraint.name

    if schema is None or constraint_name is None:
        return

    comment_map: CommentMap = reflect_all_constraint_comments(engine=engine, schema=schema)
    comment: Optional[str] = comment_map.get(schema, {}).get(table_name, {}).get(constraint_name)

    # constraint.info is "Optional[Mapping[str, Any]]"
    if not hasattr(constraint, "info"):
        constraint.info = {"comment": comment}
    else:
        constraint.info["comment"] = comment  # type: ignore

    return


@lru_cache()
def reflect_all_constraint_comments(engine, schema: str) -> CommentMap:
    """Collect a mapping of constraint comments"""

    sql = sql_text(
        """
    select
            c.relnamespace::regnamespace::text schemaname,
            c.relname tablename,
            t.conname constraintname,
            d.description comment_body
    from pg_class c
            join pg_constraint t
                    on c.oid = t.conrelid
            join pg_description d
                    on t.oid = d.objoid
                    and t.tableoid = d.classoid
    where
            c.relnamespace::regnamespace::text = :schema
    """
    )

    results = engine.execute(sql, schema=schema).fetchall()

    comment_map: CommentMap = {}

    for schema_name, table_name, constraint_name, comment in results:
        comment_map[schema_name] = comment_map.get(schema_name, {})
        comment_map[schema_name][table_name] = comment_map[schema_name].get(table_name, {})
        comment_map[schema_name][table_name][constraint_name] = comment

    return comment_map



from inspect import isclass
from typing import Optional, Tuple, Type, Union

from nebulo.env import EnvManager
from nebulo.sql.inspect import get_comment, get_foreign_key_constraint_from_relationship, get_table_name, to_table
from nebulo.sql.reflection.function import SQLFunction
from nebulo.sql.reflection.views import ViewMixin
from nebulo.sql.table_base import TableProtocol
from nebulo.text_utils import snake_to_camel, to_plural
from parse import parse
from sqlalchemy import ForeignKeyConstraint, Table
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.sql.schema import Column
from typing_extensions import Literal

ENV = EnvManager.get_environ()


class Config:

    # SQLAlchemy connection string
    CONNECTION = ENV.get("NEBULO_CONNECTION")

    # schema name
    SCHEMA = ENV.get("NEBULO_SCHEMA")

    # SQL Composite type identifier e.g. api.jwt_key
    JWT_IDENTIFIER = ENV.get("NEBULO_JWT_IDENTIFIER")

    # JWT secret key
    JWT_SECRET = ENV.get("NEBULO_JWT_SECRET")

    # Default SQL role to use for unauthenticated users
    DEFAULT_ROLE = ENV.get("NEBULO_DEFAULT_ROLE")

    # Default number of entries per API page
    DEFAULT_PAGE_SIZE = int(ENV.get("NEBULO_DEFAULT_PAGE_SIZE", 20))

    @staticmethod
    def function_name_mapper(sql_function: SQLFunction) -> str:
        """to_upper -> toUpper"""
        return snake_to_camel(sql_function.name, upper=False)

    @staticmethod
    def function_type_name_mapper(sql_function: SQLFunction) -> str:
        """to_upper -> ToUpper"""
        return snake_to_camel(sql_function.name, upper=True)

    @classmethod
    def table_type_name_mapper(cls, sqla_table: TableProtocol) -> str:
        table_comment = get_comment(sqla_table)

        # If an @name directive exists, use it
        maybe_name: Optional[str] = cls.read_name_directive(table_comment)
        if maybe_name:
            return maybe_name

        table_name = get_table_name(sqla_table)
        return snake_to_camel(table_name)

    @classmethod
    def table_name_mapper(cls, sqla_table: TableProtocol) -> str:
        """Return the type name with the first character lower case"""
        type_name = cls.table_type_name_mapper(sqla_table)
        return type_name[0].lower() + type_name[1:]

    @classmethod
    def column_name_mapper(cls, column: Column) -> str:
        """Return the type name with the first character lower case"""
        column_comment = get_comment(column)

        # If an @name directive exists, use it
        maybe_name: Optional[str] = cls.read_name_directive(column_comment)
        if maybe_name:
            return maybe_name

        column_key = column.key
        type_name = snake_to_camel(column_key)
        return type_name[0].lower() + type_name[1:]

    @staticmethod
    def read_name_directive(comment: str) -> Optional[str]:
        """Returns the name following a @name comment directive"""
        lines = comment.split("\n")
        for line in lines:
            line = line.strip()
            if line.startswith("@name"):
                line = line.lstrip("@name")
                name = line.strip()
                if name:
                    return name
        return None

    @staticmethod
    def enum_name_mapper(enum: Type[postgresql.base.ENUM]) -> str:
        return snake_to_camel((enum.name or "") + "_enum", upper=True)

    @staticmethod
    def read_foreign_key_comment_directive(comment: str) -> Tuple[Optional[str], Optional[str]]:
        """Accepts a comment from a foreign key backing a sqlalchemy RelationshipProperty and searches for
        a comment directive defining what the names for each side of the relationship should be for the local
        and remote graphql models on either side.

        Template:
            @name {local_name_for_remote} {remote_name_for_local}

        Example:
            Given:
                ALTER TABLE public.address
                ADD CONSTRAINT fk_address_to_person
                FOREIGN KEY (address_id)
                REFERENCES public.person (id);

            The comment

                COMMENT ON CONSTRAINT fk_address_to_person
                ON public.addresses
                IS E'@name Person Addresses';

            Would give:
                - The "Person" graphql model an "Addresses" field
                - The "Adress" graphql model a "Person" field
        """

        template = "@name {local_name_for_remote} {remote_name_for_local}"

        lines = comment.split("\n")
        for line in lines:
            line = line.strip()
            match = parse(template, line)
            if match is not None:
                local_name, remote_name = match.named.values()
                return (
                    local_name,
                    remote_name,
                )
        return None, None

    @classmethod
    def relationship_name_mapper(cls, relationship: RelationshipProperty) -> str:

        # Check the foreign key backing the relationship for a comment defining its name
        backing_fkey = get_foreign_key_constraint_from_relationship(relationship)
        if backing_fkey is not None:
            comment: str = get_comment(backing_fkey)
            local_name, remote_name = cls.read_foreign_key_comment_directive(comment)
            if local_name and remote_name:
                backing_fkey_parent_entity = backing_fkey.parent  # type: ignore
                fkey_parent_table: Table = to_table(backing_fkey_parent_entity)
                relationship_parent_table: Table = to_table(relationship.parent)
                if fkey_parent_table == relationship_parent_table:
                    return local_name
                return remote_name

        # No comment directive existed, reverting to default

        # Union of Mapper or ORM instance
        referred_cls = to_table(relationship.argument)
        # if hasattr(referred_cls, "class_"):
        #    referred_cls = referred_cls.class_
        # elif callable(referred_cls):
        #    referred_cls = referred_cls()

        referred_name = get_table_name(referred_cls)
        cardinal_name = to_plural(referred_name) if relationship.uselist else referred_name
        camel_name = snake_to_camel(cardinal_name, upper=False)

        relationship_name = (
            camel_name
            + "By"
            + "And".join(
                snake_to_camel(local_col.name, upper=True) + "To" + snake_to_camel(remote_col.name, upper=True)
                for local_col, remote_col in relationship.local_remote_pairs
            )
        )
        return relationship_name

    # SQL Comment Config
    @staticmethod
    def _exclude_check(
        entity: Union[TableProtocol, Column, ForeignKeyConstraint],
        operation: Union[
            Literal["read"],
            Literal["create"],
            Literal["update"],
            Literal["delete"],
            Literal["read_one"],
            Literal["read_all"],
        ],
    ) -> bool:
        """Shared SQL comment parsing logic for excludes"""
        comment: str = get_comment(entity)
        lines = comment.split("\n")
        for line in lines:
            if "@exclude" in line:
                operations_dirty = line[len("@exclude") :].split(",")
                operations_clean = {x.strip() for x in operations_dirty}
                if operation in operations_clean:
                    return True
        return False

    ###########
    # Exclude #
    ###########

    @classmethod
    def exclude_read(cls, entity: Union[TableProtocol, Column, ForeignKeyConstraint]) -> bool:
        """Should the entity be excluded from reads? e.g. entity(nodeId ...) and allEntities(...)"""
        return cls._exclude_check(entity, "read")

    @classmethod
    def exclude_read_one(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from reads? e.g. entity(nodeId ...)"""
        return any([cls._exclude_check(entity, "read_one"), cls.exclude_read(entity)])

    @classmethod
    def exclude_read_all(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from reads? e.g. allEntities(...)"""
        return any([cls._exclude_check(entity, "read_all"), cls.exclude_read(entity)])

    @classmethod
    def exclude_create(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from create mutations?"""
        # Views do not support insert
        if isclass(entity) and issubclass(entity, ViewMixin):  # type: ignore
            return True

        if isinstance(entity, Column) and entity.primary_key:
            # Primary keys must be server generated
            return True

        return cls._exclude_check(entity, "create")

    @classmethod
    def exclude_update(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from update mutations?"""

        # Views do not support updates
        if isclass(entity) and issubclass(entity, ViewMixin):  # type: ignore
            return True

        if isinstance(entity, Column) and entity.primary_key:
            # Primary keys can not be upadted via the API
            return True

        return cls._exclude_check(entity, "update")

    @classmethod
    def exclude_delete(cls, entity: Union[TableProtocol, Column]) -> bool:
        """Should the entity be excluded from delete mutations?"""

        # Views do not support deletes
        if isclass(entity) and issubclass(entity, ViewMixin):  # type: ignore
            return True

        return cls._exclude_check(entity, "delete")



import os
from functools import lru_cache
from pathlib import Path

from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles

GRAPHIQL_STATIC_PATH = Path(os.path.abspath(__file__)).parent.parent.parent.resolve() / "static"

GRAPHIQL_STATIC_FILES = StaticFiles(directory=GRAPHIQL_STATIC_PATH)


def get_graphiql_route(graphiql_path: str = "/graphiql", graphql_path: str = "/", name="graphiql") -> Mount:
    """Return mountable routes serving GraphiQL interactive API explorer

    **Parameters**

    * **graphiql_path**: _str_ = URL path to GraphiQL html page
    * **graphql_path**: _str_ =  URL path to the GraphQL route
    * **name**: _str_ =  Name for the mount
    """

    async def graphiql_endpoint(_: Request) -> HTMLResponse:
        """Return the HTMLResponse for GraphiQL GraphQL explorer configured to hit the correct endpoint"""
        html_text = build_graphiql_html(graphiql_path, graphql_path)
        return HTMLResponse(html_text)

    graphiql_html_route = Route("/", graphiql_endpoint, methods=["GET"])
    graphiql_statics = Mount("/static", GRAPHIQL_STATIC_FILES, name="static")

    return Mount(graphiql_path, routes=[graphiql_html_route, graphiql_statics], name=name)


@lru_cache()
def build_graphiql_html(
    graphiql_route_path: str,
    graphql_route_path: str,
) -> str:
    """Return the raw HTML for GraphiQL GraphQL explorer"""

    text = GRAPHIQL.replace("{{GRAPHIQL_PATH}}", graphiql_route_path)
    text = text.replace("{{REQUEST_PATH}}", f'"{graphql_route_path}"')
    return text


GRAPHIQL = """
<html>
  <head>
    <title>Nebulo GraphiQL</title>
    <link href="{{GRAPHIQL_PATH}}/static/graphiql.min.css" rel="stylesheet" />
  </head>
  <body style="margin: 0;">
    <div id="graphiql" style="height: 100vh;"></div>

    <script
      crossorigin
      src="https://unpkg.com/react/umd/react.production.min.js"
    ></script>
    <script
      crossorigin
      src="https://unpkg.com/react-dom/umd/react-dom.production.min.js"
    ></script>
    <script
      crossorigin
      src="{{GRAPHIQL_PATH}}/static/graphiql.min.js"
    ></script>

    <script>

        let graphQLFetcher = (graphQLParams, opts = { headers: {} }) => {
          return fetch(window.location.origin + {{REQUEST_PATH}}, {
            method: 'post',
            headers: Object.assign({ 'Content-Type': 'application/json', 'User-Agent': 'GraphiQL' }, opts.headers),
            body: JSON.stringify(graphQLParams),
          }).then(response => response.json());
        }

        ReactDOM.render(
          React.createElement(GraphiQL, {
            fetcher: graphQLFetcher,
            headerEditorEnabled: true,
          }),
        document.getElementById('graphiql'),
      );
    </script>
  </body>
</html>
"""



# pylint: disable=unsubscriptable-object, invalid-name
from __future__ import annotations

from functools import lru_cache
from typing import Callable, List, Optional, Union

from nebulo.sql.table_base import TableProtocol
from sqlalchemy import Column, ForeignKeyConstraint, Table
from sqlalchemy import inspect as sql_inspect
from sqlalchemy.orm import Mapper, RelationshipProperty
from sqlalchemy.sql.schema import Constraint


@lru_cache()
def get_table_name(entity: Union[Table, TableProtocol]) -> str:
    """Name of the table"""
    return str(to_table(entity).name)


@lru_cache()
def get_relationships(sqla_model: TableProtocol) -> List[RelationshipProperty]:
    """Relationships with other tables"""
    return list(sql_inspect(sqla_model).relationships)


@lru_cache()
def get_primary_key_columns(sqla_model: TableProtocol) -> List[Column]:
    """Primary key"""
    return [x for x in sqla_model.__table__.primary_key.columns]


@lru_cache()
def get_columns(sqla_model: TableProtocol) -> List[Column]:
    """Columns on the table"""
    return [x for x in sqla_model.__table__.columns]


@lru_cache()
def get_constraints(entity: Union[TableProtocol, Table]) -> List[Constraint]:
    """Retrieve constraints from a table"""
    return list(to_table(entity).constraints)


@lru_cache()
def is_nullable(relationship: RelationshipProperty) -> bool:
    """Checks if a sqlalchemy orm relationship is nullable"""
    for local_col, remote_col in relationship.local_remote_pairs:
        if local_col.nullable or remote_col.nullable:
            return True
    return False


@lru_cache()
def get_comment(entity: Union[Table, TableProtocol, Column, Constraint]) -> str:
    """Get comment on entity"""
    if isinstance(entity, TableProtocol):
        return to_table(entity).comment or ""
    elif isinstance(entity, Table):
        return entity.comment or ""
    elif isinstance(entity, Column):
        return entity.comment or ""
    elif isinstance(entity, Constraint):
        if hasattr(entity, "info"):
            return getattr(entity, "info").get("comment") or ""
        return ""
    raise ValueError("invalid entity passed to get_comment")


@lru_cache()
def get_foreign_key_constraints(entity: Union[TableProtocol, Table]) -> List[ForeignKeyConstraint]:
    """Retrieve all foreign keys associated with the table"""
    constraints = get_constraints(entity)
    fkeys = [x for x in constraints if isinstance(x, ForeignKeyConstraint)]
    return fkeys


@lru_cache()
def get_foreign_key_constraint_from_relationship(relationship: RelationshipProperty) -> Optional[ForeignKeyConstraint]:
    """Get the ForeignKeyConstraint that backs the input RelationshipProperty

    Note: the resolution method checks that the columns associated with the relationship
        match the columns associated with the foreign key constraint. If two identical
        foreign keys exist (they never should), the behavior is undefined
    """
    local_cols = list(relationship.local_columns)
    remote_cols = list(relationship.remote_side)

    # All columns associated with the relationship
    relationship_cols = set([*local_cols, *remote_cols])

    local_table: Table = local_cols[0].table
    remote_table: Table = remote_cols[0].table

    possible_fkeys = [
        *get_foreign_key_constraints(local_table),
        *get_foreign_key_constraints(remote_table),
    ]

    for fkey in possible_fkeys:

        # All local and remote columns associated with the primary key
        fkey_involved_cols = set([*list(fkey.columns), *[x.column for x in fkey.elements]])

        if relationship_cols == fkey_involved_cols:
            return fkey
    return None


@lru_cache()
def to_table(entity: Union[Table, TableProtocol, Mapper, Callable[[], Union[TableProtocol, Table]]]) -> Table:
    """Coerces Table and ORM Table to Table"""
    if isinstance(entity, Table):
        return entity
    elif isinstance(entity, Mapper):
        return to_table(entity.class_)
    elif isinstance(entity, TableProtocol):
        return entity.__table__
    elif callable(entity):
        return to_table(entity())
    raise NotImplementedError("invalid input type")



import typing

from sqlalchemy import Text, func, literal, select
from sqlalchemy.sql.selectable import Select


def build_claims(jwt_claims: typing.Dict[str, typing.Any], default_role: typing.Optional[str] = None) -> Select:
    """Emit statement to set 'jwt.claims.<key>' for each claim in claims dict
    and a 'role'
    """
    # Setting local variables an not be done in prepared statement
    # since JWT claims are signed, literal binds should be ok

    # Keys with special meaning that should not be prefixed with 'jwt.claims'
    role_key = "role"
    special_keys = {role_key, "statement_timeout"}

    # Set all jwt.claims.*
    claims = [
        func.set_config(
            literal("jwt.claims.").op("||")(func.cast(claim_key, Text())),
            func.cast(str(claim_value), Text()),
            True,
        )
        for claim_key, claim_value in jwt_claims.items()
        if claim_key not in special_keys
    ]

    # Set role to default if exists and not provided by jwt
    if role_key not in jwt_claims and default_role is not None:
        jwt_claims[role_key] = default_role

    # Set keys with special meaning to postgres
    for key in special_keys:
        if key in jwt_claims:
            claims.append(
                func.set_config(
                    func.cast(key, Text()),
                    func.cast(str(jwt_claims[key]), Text()),
                    True,
                )
            )

    return select(claims)
